#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

static const BulletMLUserFunc userFunc;

typedef struct _BULLET {
	const char* fname;
	BulletMLNode* bulletml;
} BULLET;

static BULLET bullet_tbl[BULLET_COUNT] = {
	{ "bullet01.xml" },
	{ "bullet02.xml" },
};

void
bulletINIT()
{
	const void* fpk = resource_address(APP_FNAME);
	int i;
	int size;
	const char* xml;
	char* buf;

	for(i = 0; i < BULLET_COUNT; i++) {
		xml = fpk_address(fpk, bullet_tbl[i].fname, &size);
		buf = calloc(size + 1/*nul*/, 1);
		if(!buf) DIE();
		memcpy(buf, xml, size);
		bullet_tbl[i].bulletml = BulletML_load(buf);
		free(buf);
	}
}

void
bulletSet(TSK* tsk, int no)
{
	bulletDel(tsk);
	tsk->bullet = BulletMLBullet_new(&userFunc, tsk, bullet_tbl[no].bulletml, NULL);
}

void
bulletDel(TSK* tsk)
{
	BulletMLBullet_delete(tsk->bullet/*NULLOK*/);
	tsk->bullet = NULL;
}

int
bulletRun(TSK* tsk)
{
	if(tsk->bullet) {
		if(!BulletMLBullet_run(tsk->bullet)) {
			bulletDel(tsk);
		}
	}
	return (int)tsk->bullet;
}

/****************************************************************************
 *	
 ****************************************************************************/

#define ROTVAL	fld(180.0 / M_PI)

static float getShipDirection(TSK* tsk) {
	float px;
	float py;
	float dir;
	if(ship_tsk) {
		px = tsk->px - ship_tsk->px;
		py = tsk->py - ship_tsk->py;
		dir = atan2(px, py);
	} else {
		dir = 0;
	}
	return dir;
}

/*--------------------------------------------------------------------------*/

static void createBullet(void* userData, fixed dir, fixed spd, BulletMLNode* bulletNode, const fixed params[/*BULLETML_MAX_PARAMS*/]) {
	TSK* tsk = userData;
	TSK* etsk = setTSK(GROUP_ESHOT, TSKeshot);
	etsk->px = tsk->px;
	etsk->py = tsk->py;
#if 0
	etsk->bullet_dir = fdiv(dir, ROTVAL);
	etsk->bullet_spd = spd;
#else
	etsk->bullet_dir = fst(fdiv(dir, ROTVAL));
	etsk->bullet_spd = fst(spd);
#endif
	etsk->bullet_accx = tsk->bullet_accx;
	etsk->bullet_accy = tsk->bullet_accy;
	etsk->bullet = BulletMLBullet_new(&userFunc, etsk, bulletNode, params);
}
static void setDirection(void* userData, fixed dir) {
	TSK* tsk = userData;
#if 0
	tsk->bullet_dir = fdiv(dir, ROTVAL);
#else
	tsk->bullet_dir = fst(fdiv(dir, ROTVAL));
#endif
}
static void setSpeed(void* userData, fixed spd) {
	TSK* tsk = userData;
#if 0
	tsk->bullet_spd = spd;
#else
	tsk->bullet_spd = fst(spd);
#endif
}
static void setAccelH(void* userData, fixed accH) {
	TSK* tsk = userData;
#if 0
	tsk->bullet_accx = accH;
#else
	tsk->bullet_accx = fst(accH);
#endif
}
static void setAccelV(void* userData, fixed accV) {
	TSK* tsk = userData;
#if 0
	tsk->bullet_accy = accV;
#else
	tsk->bullet_accy = fst(accV);
#endif
}
static void vanish(void* userData) {
	TSK* tsk = userData;
	tsk->step = -1;
}
static fixed getRand(void* userData) {
	return FRND(seed);
}
static fixed getRank(void* userData) {
	TSK* tsk = userData;
#if 0
	return tsk->rank;
#else
	return fld(tsk->rank);
#endif
}
static fixed getAimDirection(void* userData) {
	TSK* tsk = userData;
#if 0
	return fmul(getShipDirection(tsk), ROTVAL);
#else
	return fmul(fld(getShipDirection(tsk)), ROTVAL);
#endif
}
static fixed getDirection(void* userData) {
	TSK* tsk = userData;
#if 0
	return fmul(tsk->bullet_dir, ROTVAL);
#else
	return fmul(fld(tsk->bullet_dir), ROTVAL);
#endif
}
static fixed getSpeed(void* userData) {
	TSK* tsk = userData;
#if 0
	return tsk->bullet_spd;
#else
	return fld(tsk->bullet_spd);
#endif
}
static fixed getAccelH(void* userData) {
	TSK* tsk = userData;
#if 0
	return tsk->bullet_accx;
#else
	return fld(tsk->bullet_accx);
#endif
}

static fixed
getAccelV(void* userData)
{
	TSK* tsk = userData;
#if 0
	return tsk->bullet_accy;
#else
	return fld(tsk->bullet_accy);
#endif
}

/*--------------------------------------------------------------------------*/

static const BulletMLUserFunc userFunc = {
	.createBullet    = createBullet,
	.setDirection    = setDirection,
	.setSpeed        = setSpeed,
	.setAccelH       = setAccelH,
	.setAccelV       = setAccelV,
	.vanish	         = vanish,
	.getRand         = getRand,
	.getRank         = getRank,
	.getAimDirection = getAimDirection,
	.getDirection    = getDirection,
	.getSpeed        = getSpeed,
	.getAccelH       = getAccelH,
	.getAccelV       = getAccelV,
};
