#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

static void TSKparticle(TSK* tsk);
static void TSKparticleDraw(TSK* tsk);

void
effSetParticle(TSK* tsk, int cnt)
{
	int i;
	TSK* etsk;

	for(i = 0; i < cnt; i++) {
		etsk = setTSK(GROUP_EFFECT, TSKparticle);
		etsk->px = tsk->px;
		etsk->py = tsk->py;
	}
}

static void
TSKparticle(TSK* tsk)
{
	int i;
	float rot;
	float rad;
	float poly[3][2];

	switch(tsk->step) {
	case 0:
		for(i = 0; i < 3; i++) {
			switch(i) {
			case 0:
				poly[i][0] = -((rand() % 4096) / 1024.0f + 1.0f);
				poly[i][1] = +((rand() % 4096) / 1024.0f + 1.0f);
				break;
			case 1:
				poly[i][0] = +((rand() % 2048) / 1024.0f - 1.0f);
				poly[i][1] = -((rand() % 4096) / 1024.0f + 1.0f);
				break;
			default:
				poly[i][0] = +((rand() % 4096) / 1024.0f + 1.0f);
				poly[i][1] = +((rand() % 4096) / 1024.0f + 1.0f);
				break;
			}
		}
		setForm(tsk, poly, 3, 1.0f);
		tsk->fp_draw = TSKparticleDraw;
		rot = (rand() % 10000) / 10000.0f * M_PI2;
		rad = rand() % 256 + 256.0f;
		tsk->tx = tsk->px + sin(rot) * rad;
		tsk->ty = tsk->py + cos(rot) * rad;
		tsk->wait = 60;
		tsk->step++;
		break;

	case 1:
		tsk->px += (tsk->tx - tsk->px) / 120.0f;
		tsk->py += (tsk->ty - tsk->py) / 120.0f;
		if((tsk->rot += M_PI / 30) >= M_PI) tsk->rot -= M_PI2;
		if(--tsk->wait <= 0) {
			tsk->step = -1;
		}
		break;

	default:
		clrTSK(tsk);
		break;
	}
}

static void
TSKparticleDraw(TSK* tsk)
{
	int c;

	c = tsk->wait / 20;	/* 60,59->0 => 3,2->0 */
	if(c < 0) c = 0;
	if(c > 2) c = 2;
	c = 2 - c;

	drawLineLoop(tsk, 0, tsk->body_len, c);
}

/****************************************************************************
 *	
 ****************************************************************************/

static void TSKBrokenBody(TSK* tsk);
static void TSKBrokenBodyDraw(TSK* tsk);

void
effSetBrokenBody(TSK* tsk, const float poly[/*cnt*/][2], int cnt, float scale)
{
	int i;
	float cx;
	float cy;
	float tmp[4/**/][2];
	TSK* etsk;

	if(cnt > ARRAY_SIZE(tmp)) DIE();

	cx = 0.0f;
	cy = 0.0f;
	for(i = 0; i < cnt; i++) {
		tmp[i][0] = poly[i][0];
		tmp[i][1] = poly[i][1];
		cx += tmp[i][0];
		cy += tmp[i][1];
	}
	cx /= cnt;
	cy /= cnt;
	for(i = 0; i < cnt; i++) {
		tmp[i][0] -= cx;
		tmp[i][1] -= cy;
	}

	etsk = setTSK(GROUP_EFFECT, TSKBrokenBody);
	etsk->px = tsk->px;
	etsk->py = tsk->py;
	setForm(etsk, tmp, cnt, scale);
}

static void
TSKBrokenBody(TSK* tsk)
{
	float rot;
	float rad;

	switch(tsk->step) {
	case 0:
		tsk->fp_draw = TSKBrokenBodyDraw;
		rot = (rand() % 10000) / 10000.0f * M_PI2;
		rad = rand() % 256 + 256.0f;
		tsk->tx = tsk->px + sin(rot) * rad;
		tsk->ty = tsk->py + cos(rot) * rad;
		tsk->rot_add = (rand() % 30) - 15.0f;	/* -15`+15 */
		if(tsk->rot_add < 0) {
			tsk->rot_add -= 15.0f;		/* -30`-15 */
		} else {
			tsk->rot_add += 15.0f;		/* +15`+30 */
		}
		tsk->rot_add = M_PI / tsk->rot_add;
		tsk->wait = 60;
		tsk->step++;
		break;

	case 1:
		tsk->px += (tsk->tx - tsk->px) / 120.0f;
		tsk->py += (tsk->ty - tsk->py) / 120.0f;
		tsk->rot += tsk->rot_add;
		if(tsk->rot < -M_PI) tsk->rot += M_PI2;
		if(tsk->rot >= M_PI) tsk->rot -= M_PI2;
		if(--tsk->wait <= 0) {
			tsk->step = -1;
		}
		break;

	default:
		clrTSK(tsk);
		break;
	}
}

static void
TSKBrokenBodyDraw(TSK* tsk)
{
	int c;

	c = tsk->wait / 20;	/* 60,59->0 => 3,2->0 */
	if(c < 0) c = 0;
	if(c > 2) c = 2;
	c = 2 - c;

	drawLineLoop(tsk, 0, tsk->body_len, c);
}

/****************************************************************************
 *	
 ****************************************************************************/

static void TSKBrokenLine(TSK* tsk);
static void TSKBrokenLineDraw(TSK* tsk);

#define MAX_LINES	1000
static int now_lines;

void
effSetBrokenLine(TSK* tsk, const float poly[/*cnt*/][2], int cnt, float scale)
{
	int i;
	float x1;
	float y1;
	float x2;
	float y2;
	float cx;
	float cy;
	float tmp[2][2];
	TSK* etsk;

	x2 = poly[cnt - 1][0];
	y2 = poly[cnt - 1][1];
	for(i = 0; i < cnt && now_lines < MAX_LINES; i++, now_lines++) {
		x1 = x2;
		y1 = y2;
		x2 = poly[i][0];
		y2 = poly[i][1];
		cx = (x1 + x2) / 2.0f;
		cy = (y1 + y2) / 2.0f;
		tmp[0][0] = x1 - cx;
		tmp[0][1] = y1 - cy;
		tmp[1][0] = x2 - cx;
		tmp[1][1] = y2 - cy;

		etsk = setTSK(GROUP_EFFECT, TSKBrokenLine);
		etsk->px = tsk->px;
		etsk->py = tsk->py;
		setForm(etsk, tmp, 2, scale);
	}
}

static void
TSKBrokenLine(TSK* tsk)
{
	float rot;
	float rad;

	switch(tsk->step) {
	case 0:
		tsk->fp_draw = TSKBrokenLineDraw;
		rot = (rand() % 10000) / 10000.0f * M_PI2;
		rad = rand() % 256 + 256.0f;
		tsk->tx = tsk->px + sin(rot) * rad;
		tsk->ty = tsk->py + cos(rot) * rad;
		tsk->rot_add = (rand() % 30) - 15.0f;	/* -15`+15 */
		if(tsk->rot_add < 0) {
			tsk->rot_add -= 15.0f;		/* -30`-15 */
		} else {
			tsk->rot_add += 15.0f;		/* +15`+30 */
		}
		tsk->rot_add = M_PI / tsk->rot_add;
		tsk->wait = 60;
		tsk->step++;
		break;

	case 1:
		tsk->px += (tsk->tx - tsk->px) / 120.0f;
		tsk->py += (tsk->ty - tsk->py) / 120.0f;
		tsk->rot += tsk->rot_add;
		if(tsk->rot < -M_PI) tsk->rot += M_PI2;
		if(tsk->rot >= M_PI) tsk->rot -= M_PI2;
		if(--tsk->wait <= 0) {
			tsk->step = -1;
		}
		break;

	default:
		now_lines--;
		clrTSK(tsk);
		break;
	}
}

static void
TSKBrokenLineDraw(TSK* tsk)
{
	int c;

	c = tsk->wait / 20;	/* 60,59->0 => 3,2->0 */
	if(c < 0) c = 0;
	if(c > 2) c = 2;
	c = 2 - c;

	drawLines(tsk, 0, tsk->body_len, c);
}

