#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

#define SSHOT_MAX	12
#define SSHOT_WAIT_MAX	3
static int sshot_cnt;
static int sshot_wait;

/****************************************************************************
 *	
 ****************************************************************************/

static void TSKshipInt(TSK* tsk, TSK* etsk);
static void TSKshipDraw(TSK* tsk);
static void TSKsshot(TSK* tsk);
static void TSKsshotInt(TSK* tsk, TSK* etsk);
static void TSKsshotDraw(TSK* tsk);

#define SHIP_SCALE	2.0f
static const float ship_poly[][2] = {
	{  -4.0f, -4.0f },
	{  +0.0f, -8.0f },
	{  +4.0f, -4.0f },
	{  +0.0f, +8.0f },
	//
	{  -2.0f, -2.0f },
	{  -6.0f, -2.0f },
	{ -10.0f,-10.0f },
	{  -2.0f, -6.0f },
	//
	{  +2.0f, -2.0f },
	{  +6.0f, -2.0f },
	{ +10.0f,-10.0f },
	{  +2.0f, -6.0f },
};

#define SHIP_SRATE	(1.0f / M_SQRT2)
#define SHIP_SX1	8.0f
#define SHIP_SY1	8.0f
#define SHIP_SX2	(SHIP_SX1 * SHIP_SRATE)
#define SHIP_SY2	(SHIP_SY1 * SHIP_SRATE)
static const float ship_move[][2] = {
	{     +0.0f,    +0.0f },	/* -- -- -- -- */
	{ -SHIP_SX1,    +0.0f },	/* -- -- -- RI */
	{ +SHIP_SX1,    +0.0f },	/* -- -- LF -- */
	{     +0.0f,    +0.0f },	/* -- -- LF RI */
	{     +0.0f,+SHIP_SY1 },	/* -- DN -- -- */
	{ -SHIP_SX2,+SHIP_SY2 },	/* -- DN -- RI */
	{ +SHIP_SX2,+SHIP_SY2 },	/* -- DN LF -- */
	{     +0.0f,    +0.0f },	/* -- DN LF RI */
	{     +0.0f,-SHIP_SY1 },	/* UP -- -- -- */
	{ -SHIP_SX2,-SHIP_SY2 },	/* UP -- -- RI */
	{ +SHIP_SX2,-SHIP_SY2 },	/* UP -- LF -- */
	{     +0.0f,    +0.0f },	/* UP -- LF RI */
	{     +0.0f,    +0.0f },	/* UP DN -- -- */
	{     +0.0f,    +0.0f },	/* UP DN -- RI */
	{     +0.0f,    +0.0f },	/* UP DN LF -- */
	{     +0.0f,    +0.0f },	/* UP DN LF RI */
};

void
TSKship(TSK* tsk)
{
	int ship_pad = joy & (PAD_UP|PAD_DN|PAD_LF|PAD_RI|PAD_A);
	int ship_dir = joy & (PAD_UP|PAD_DN|PAD_LF|PAD_RI);
	TSK* etsk;

	if((g_step == GSTEP_RESULT) && (tsk->step != 10)) {
		ship_tsk = NULL;
		tsk->fp_int = TSKshipInt;
		tsk->fp_draw = TSKshipDraw;
		tsk->step = 10;
	}

	switch(tsk->step) {
	case 0:
		ship_tsk = tsk;
		sshot_cnt = 0;
		sshot_wait = 0;
		tsk->cx = 2.0f;
		tsk->cy = 2.0f;
		tsk->cnt = 60;
		tsk->wait = 0;
		tsk->step++;
		/*FALLTHRU*/

	case 1:
		if(!tsk->wait) {
			tsk->fp_draw = TSKshipDraw;
			tsk->cen_x =      0.0f + 32.0f;
			tsk->cen_y = SCREEN_MY - 16.0f;
			tsk->rad_x =  32.0f;
			tsk->rad_y = 128.0f;
			tsk->ang = M_PI2 / 5.0f;
			tsk->wait = 30;
		}
		tsk->px = tsk->cen_x + sin(tsk->ang) * tsk->rad_x;
		tsk->py = tsk->cen_y + cos(tsk->ang) * tsk->rad_y;
		if((tsk->ang += M_PI / 30.0f) >= M_PI) tsk->ang -= M_PI2;
		if(--tsk->wait <= 0) {
			tsk->step++;
		}
		break;

	case 2:
		if(--tsk->cnt <= 0) {
			tsk->fp_int = TSKshipInt;
			tsk->step++;
		}
		/*FALLTHRU*/

	case 3:
		tsk->vx = ship_move[ship_dir][0];
		tsk->vy = ship_move[ship_dir][1];
		tsk->px += tsk->vx;
		tsk->py += tsk->vy;
		if(tsk->px < -SHIP_AREAMAX_X) tsk->px = -SHIP_AREAMAX_X;
		if(tsk->px > +SHIP_AREAMAX_X) tsk->px = +SHIP_AREAMAX_X;
		if(tsk->py < -SHIP_AREAMAX_Y) tsk->py = -SHIP_AREAMAX_Y;
		if(tsk->py > +SHIP_AREAMAX_Y) tsk->py = +SHIP_AREAMAX_Y;
		if(sshot_wait) {
			sshot_wait--;
		} else {
			if((ship_pad & PAD_A) && (sshot_cnt < SSHOT_MAX)) {
				sshot_cnt++;
				sshot_wait = SSHOT_WAIT_MAX;
				etsk = setTSK(GROUP_SSHOT, TSKsshot);
				etsk->px = tsk->px;
				etsk->py = tsk->py;
				playSND(SND_SESSHOT);
			}
		}
		break;

	case 10:
		/** no job **/
		break;

	case 100:
		if(--tsk->wait <= 0) {
			tsk->cnt = 60;
			tsk->wait = 0;
			tsk->step = 1;
		}
		break;

	default:
		clrTSK(tsk);
		break;
	}
}

static void
TSKshipInt(TSK* tsk, TSK* etsk)
{
	playSND(SND_SESBRK);
	effSetBrokenBody(tsk, ship_poly + 0, 4, 3.0f);
	effSetBrokenBody(tsk, ship_poly + 4, 4, 3.0f);
	effSetBrokenBody(tsk, ship_poly + 8, 4, 3.0f);
	effSetBrokenLine(tsk, ship_poly + 0, 4, 3.0f);
	effSetBrokenLine(tsk, ship_poly + 4, 4, 3.0f);
	effSetBrokenLine(tsk, ship_poly + 8, 4, 3.0f);
	tsk->fp_int = NULL;
	tsk->fp_draw = NULL;
	tsk->wait = 30;
	tsk->step = 100;
	brk_ship++;
	score -= 10;
	if(score < 0) score = 0;
}

static void
TSKshipDraw(TSK* tsk)
{
	if(tsk->fp_int || (now & 4)) {
		int x = _getPointX(-tsk->px);
		int y = _getPointY(-tsk->py);
		sprite_draw(x, y, SPR_SHIP, DRW_NOMAL);
	}
}

/****************************************************************************
 *	
 ****************************************************************************/

static void
TSKsshot(TSK* tsk)
{
	switch(tsk->step) {
	case 0:
		tsk->fp_int = TSKsshotInt;
		tsk->fp_draw = TSKsshotDraw;
		tsk->cx = 6.0f;
		tsk->cy = 6.0f;
		tsk->step++;
		break;

	case 1:
		tsk->py -= 12.0f;
		if(tsk->py < -SHIP_AREAMAX_Y) {
			tsk->step = -1;
		}
		break;

	default:
		sshot_cnt--;
		clrTSK(tsk);
		break;
	}
}

static void
TSKsshotInt(TSK* tsk, TSK* etsk)
{
	tsk->step = -1;
}

static void
TSKsshotDraw(TSK* tsk)
{
	int x = _getPointX(-tsk->px);
	int y = _getPointY(-tsk->py);
	sprite_draw(x, y, SPR_SSHOT, DRW_NOMAL);
}

